
#ifndef _GPEMAP_HEADER
#define _GPEMAP_HEADER

#include <tchar.h>
#include <windows.h>


/*

	*** EXAMPLE USAGE ***

	1) To get the name of a function :
		{
			GPEMap					tmpMap;
			GPEMap::CodeSourceData	tmpData;
			if (tmpMap.GetCodeSource(_SomeFunction,&tmpData))
			{
				//	tmpData now contains the filename, linenumber and name of _SomeFunction...
			}
		}

	2) o 

*/

//
//	Testing...
//

//	Comment this out to remove the testing functions
#define _TESTING

#ifdef _TESTING

	extern void _TestGPEMap(const TCHAR * pPEFile);

#endif


//
//	Forward Struct Declarations
//


struct ImageDebugDirectory;
struct CVSignature;
struct CVSubSection;


//
//	Classes
//

class GPEMap
{
	public :

		//
		//	Enums
		//

		enum
		{
			eMaxSourceFileName	= 512,				//	Max. size of the source filename
			eMaxFncName			= 256				//	Max. size of the enclosing function name
		};

		//
		//	Structs
		//

		struct CodeSourceData
		{
			TCHAR	m_File[eMaxSourceFileName];		//	Source file containing the code
			int		m_Line;							//	Source line the code appears on
			TCHAR	m_Fnc[eMaxFncName];				//	Name of the function the code appears in
		}; //CodeSourceData

	private :

		//
		//	Members
		//

		bool						m_bOK;				//	true if all initialization and processing went ok

		HANDLE						m_hFile;			//	Actual PE file handle
		HANDLE						m_hMapFile;			//	Memory mapped PE file

		const unsigned char *		m_pMapAddress;		//	Base address of the mapped file

		unsigned int				m_SystemPageSize;	//	Size of a page on this system (4096)

		//
		//	COFF Debug Members
		//

		ImageDebugDirectory *		m_pCOFFDbgDir;		//	Pointer to the COFF data directory in the mapped file
		//	TODO...

		//
		//	CV Debug Members
		//

		const ImageDebugDirectory *	m_pCVDbgDir;		//	Pointer to the CV data directory in the mapped file
		const unsigned char *		m_pLFO;				//	LFO CV pointer (base for all cv debug info)
		const CVSignature *			m_pCVSig;			//	CV signature
		const CVSubSection *		m_pCVGlobalSym;		//	Pointer to the CV global symbol table
		const CVSubSection * *		m_pCVSrcMods;		//	Array of CV source module section pointers
		const CVSubSection * *		m_pCVAlignSym;		//	Alignment symbol tables for all modules
		int							m_NumCVSrcMods;		//	Number of pointers in m_pCVSrcMods
		int							m_NumCVAlignSyms;	//	Number of alignment symbol subsections in m_pCVAlignSym




	public :

		//	Constructor - sets up a file mapping of the current module
					GPEMap();

		//	Constructor - sets up a file mapping of the specified PE file
					GPEMap(const TCHAR * pPEFileToMap);

	private :

		//	Shared constructor initialization function
		void		Init();

		//	PE File management
		bool		OpenPEFile(const TCHAR * pPEFileToMap);
		bool		ProcessPEFile();
		void		ClosePEFile();

		//	CV and COFF debug directory processing
		void		ProcessCVDbgDir(const ImageDebugDirectory * pDebugDir);
		void		ProcessCOFFDbgDir(const ImageDebugDirectory * pDebugDir);

		//	Error output (to message box)
		void		OutputErrorF(const TCHAR * pError,...) const;

		//	Success check - returns true if the GPEMap constructor successfully created the file mapping
		bool		IsOK() const		{ return m_bOK;					}

		bool		UsingCOFF() const	{ return m_pCOFFDbgDir != 0;	}
		bool		UsingCV() const		{ return m_pCVDbgDir != 0;		}


		//
		//	CodeView Methods
		//

		//	CodeView signature checker (may change map file from exe to pdb)
		bool		CheckCVSig(const unsigned int Sig);
		
		//	CodeView source lookup - note that the symbol lookup can be adapted to lookup the names
		//	of variables too (see notes in source)
		bool		GetCVCodeSource(const void * pAddr,CodeSourceData * pData) const;
		bool		CVLookupAddrSym(const unsigned int AddrUInt,CodeSourceData * pData) const;
		bool		CVLookupAddrSrc(const unsigned int AddrUInt,CodeSourceData * pData) const;


		//
		//	COFF Methods
		//

		//	TODO...



	public :

		//	Code source lookup
		bool		GetCodeSource(const void * pAddr,CodeSourceData * pData) const;

		//	Stack walker code source lookup (use _WalkTheStack to build up a list of code pointers for
		//	this method to interpret)
		bool		GetStackWalkCodeSource(const int		NumCalls,
										   const void * *	pStackAddrs,
										   CodeSourceData *	pData) const;

		//	Destructor - kills the file mapping
					~GPEMap();


		//
		//	Static Methods
		//
		
		//	Bonus fast stack walker - use to format pStackAddrs for a call to GetStackWalkCodeSource. Note
		//	that this assumes a standard stack frame format that uses EBP as a frame pointer. I think there
		//	are some optimizations which omit the frame pointer - check FPO_DATA in MSDN for details on 
		//	how to fixme
		static bool	_WalkTheStack(const int NumCallsToCheck,const int NumCallsToSkip,void * * pStackAddrs);

}; //GPEMap


#endif
